{-# LANGUAGE GeneralizedNewtypeDeriving #-}
module Main where 

import IO
import Data.Char
import qualified Data.Map as M
import Control.Monad.State
import Control.Monad

import Parser.Abs
import Parser.ErrM
import Parser.Par

import Backend.Abs
import Backend.Print

import GenErr
import Data()
import Options
import VarState
import GeneratorState
import ChanState
import Includes
import Help

transProgramSpec :: ProgramSpec -> GenErr BEProgram
transProgramSpec x = case x of
  Program funcdecls  -> do
    mapM transGlobDecl funcdecls
    createMainProc
    createChanDep
    getProgram

createMainProc :: GenErr()
createMainProc = do
    startProc <- helpCreateNewProcPure "Main" [genNull, genNull, genNull]
    initChans <- createChanInitFunc
    freeChans <- createChanFreeFunc
    let
      initQ = helpCallFuncStr genInitQueue []
      initWorker = helpCallFuncStr genInitWorkers []
      waitWorker = helpCallFuncStr genWaitWorkers [] in  
      addToSource $ helpCreateFuncStr helpCreateIntType  "main" [] 
        $ BECompound [initChans, initQ, startProc, initWorker, waitWorker, freeChans]

createChanInitFunc :: GenErr BEStatement
createChanInitFunc = do
    let f c = helpAssignVarFunc c genChanInit []
    exprs <- getChanList
    let stmts = map f exprs
    addToSource $ helpCreateFuncStr helpCreateVoidType genChanInitName [] $ BECompound stmts
    addToHeader $ helpCreateFuncProtoStr helpCreateVoidType genChanInitName []
    return $ helpCallFuncStr genChanInitName []

createChanFreeFunc :: GenErr BEStatement
createChanFreeFunc = do
    let f c = helpCallFuncExp genChanFree [c]
    exprs <- getChanList
    let stmts = map f exprs
    addToSource $ helpCreateFuncStr helpCreateVoidType genChanFreeName [] $ BECompound stmts
    addToHeader $ helpCreateFuncProtoStr helpCreateVoidType genChanFreeName []
    return $ helpCallFuncStr genChanFreeName []

transGlobDecl :: GlobDecl -> GenErr ()
transGlobDecl x  = case x of 
    GlobalChannelDecl chan -> transChan chan
    GlobalProcessDecl capident deadline block -> do
        prevCounters <- getCounters
        resetCounters
        (proto, func) <- parseProcDecl capident deadline block 
        addToHeader proto
        addToSource func
        setCounters  prevCounters

parseProcDecl :: CapIdent -> DeadLine -> SingleBlock -> GenErr (BEGlobal, BEGlobal)
parseProcDecl capident deadline singleblock = do
  dl <- transDeadLine deadline
  let types = helpCreateEnumType genProcEnum 
  let stackDecl =  (helpCreateStructPointerType genStackStruct, genStack)
  let procDecl =  (helpCreateStructPointerType genProcStruct, genProc)
  ident <- transCapIdent capident
  setCurrentProc ident
  let proto = helpCreateFuncProto types ident [stackDecl, procDecl]
  if ident == genMain
    then do 
      comp <- transMainProcBlock singleblock
      createProcVarDeadline capident dl  >>= \d -> addToSource d
      createProcVarCount capident >>= \var -> addToSource var 
      let func = helpCreateFunc types ident [stackDecl, procDecl] comp
      return (proto, func)
    else do
      comp <- transProcBlock singleblock
      let func = helpCreateFunc types ident [stackDecl, procDecl] comp
      createProcVarDeadline capident dl  >>= \d -> addToSource d
      createProcVarCount capident >>= \var -> addToSource var 
      return (proto, func)

createProcVarCount :: CapIdent -> GenErr BEGlobal
createProcVarCount x = case x of 
  CapIdent capIdent -> do
    let ident = genProcessVarName $ capIdent
    decl <- declVarCountSemicolon $ LowIdent ident
    return $ BEGlobalDecl $ decl

createProcVarDeadline :: CapIdent -> Integer -> GenErr BEGlobal
createProcVarDeadline str deadline = case str of
    CapIdent capIdent -> do 
       setDeadline deadline
       let ident =  genDeadline capIdent
       decl <- declVarIntSemicolon (LowIdent ident) $  fromInteger deadline
       return $ BEGlobalDecl $ decl 

transMainProcBlock :: SingleBlock -> GenErr BECompound 
transMainProcBlock block = transPureProcBlock block genStatusMainDone

transProcBlock :: SingleBlock -> GenErr BECompound
transProcBlock block = transPureProcBlock block genStatusDone

transPureProcBlock :: SingleBlock -> String -> GenErr BECompound
transPureProcBlock x endState = case x of
  SingleBlock vars procs  -> do
    upVarScope
    mapM transVar vars
    stComp <- transStmtCompound procs
    let comp =  BEStatementCompound $ stComp 
    initCo <- createCoRoutine [comp] endState
    stInit <- initProc
    fin <- finalizeProc
    downVarScope
    return $ BECompound $ stInit ++ [initCo] ++ fin

initProc :: GenErr [BEStatement]
initProc =  return $ varInit ++ stackInit ++ stackParent where 
    varInit = helpDeclAssignVarVar genProcStatus genStatusRunning 
        $ BEPrimitiveEnum (BEIdent genProcEnum)
    stackInit = helpDeclAssignPointVarFunc genStackVar genReadStack [genStack] BEPrimitiveInt
    stackParent = helpDeclAssignVarPoint (BEPrimitiveStruct (BEIdent genStackStructPoint))
                    genParentStack genStack genParentStack

finalizeProc :: GenErr [BEStatement]
finalizeProc = let 
  procFinLabel = BELabel $ BEIdent genFinalizeLabel
  procWritSt = helpCallFuncStr genWriteStack [genStack, genStackVar]
  procReturn = BEReturnValue $ BEEVar $ BEIdent genProcStatus in 
  return [procFinLabel, procWritSt, procReturn]


transSingleBlock :: SingleBlock -> GenErr BECompound
transSingleBlock x = case x of
  SingleBlock vars comps -> do
    mapM transVar vars
    compound  <- transStmtCompound comps 
    return $ compound

transStmtCompound :: StmtCompound -> GenErr BECompound
transStmtCompound x = case x of
  StmtOrder order stmtBlock -> case order of
    OrderPar -> createParBlock stmtBlock
    OrderSeq -> createSeqBlock stmtBlock
  StmtFor order lowident assignop initValue range stmtBlock -> case order of 
    OrderPar -> createParFor lowident initValue assignop range stmtBlock
    OrderSeq -> createSeqFor lowident initValue assignop range stmtBlock

createParBlock :: StmtBlock -> GenErr BECompound
createParBlock x = case x of 
    StmtBlock stmts -> do
        let stComp = map (\st -> StmtBlock [st]) stmts
        compounds <- mapM createProc stComp
        let statements = map (\cm -> BEStatementCompound cm) compounds 
        deSched <- createDeSchedule genStatusWaiting
        return $ BECompound $ statements ++ [deSched, BEEmptyStatement]

createSeqBlock :: StmtBlock -> GenErr BECompound
createSeqBlock stmtBlock = transStmtBlock stmtBlock 

createParFor :: LowIdent -> Integer -> AssignOp -> Integer -> StmtBlock -> GenErr BECompound
createParFor lowident initValue assignop range stmtblock = do 
    for <- transFor lowident initValue assignop range createProc stmtblock
    deSched <- createDeSchedule genStatusWaiting
    return $ BECompound $ [BEStatementCompound for, deSched, BEEmptyStatement]

createSeqFor :: LowIdent -> Integer -> AssignOp -> Integer -> StmtBlock -> GenErr BECompound
createSeqFor lowident initValue assignop range stmtblock = do 
  transFor lowident initValue assignop range transStmtBlock stmtblock

createProc :: StmtBlock -> GenErr BECompound 
createProc stmtblock =  do
    procName <- genProcessName
    deadline <- getDeadline
    singleblock <- transStmtBlockToSingleBlock stmtblock
    transGlobDecl $ GlobalProcessDecl 
                    (CapIdent procName) (DeadLineNs deadline) 
                    singleblock
    newProcFunc <- helpCreateNewProcChanAndProcDep procName chanMapContainsProc
    return $ BECompound $    
            [ helpCallFuncStr genWriteStack [genStack, genStackVar],
              newProcFunc
            ]    
 
transFor :: LowIdent -> Integer -> AssignOp -> Integer -> (StmtBlock -> GenErr BECompound) -> 
    StmtBlock -> GenErr BECompound 
transFor lowident start _ range f block = do
  declVar lowident 
  var <- transLowIdentToStackVar lowident
  ident <- transLowIdent lowident 
  setCurrentLoop (BEEVar ident) (fromInteger start) $ fromInteger range
  comp <- f block
  resetCurrentLoop 
  return $ BECompound $ [helpCreateFor var start range comp]

createCoRoutine :: [BEStatement] -> String ->  GenErr BEStatement
createCoRoutine stmts endStatus = do
  let caseLabel = BECaseLabel $ BEEConst $ BEConstLiteralInt $ 0
  let procStatus = helpAssignVarVar genProcStatus endStatus
  return $ BESwitch (BEEVar (BEIdent genStackPc)) $ BECompound $ 
    [caseLabel] ++ stmts ++ [procStatus]

declVarCountSemicolon :: LowIdent -> GenErr BESemicolonDecl
declVarCountSemicolon lowIdent = do
    identifier <- transLowIdent lowIdent 
    varPos <- getVarCount
    addToLocal (BEEVar identifier) varPos
    declVarIntSemicolon lowIdent varPos

declVarIntSemicolon :: LowIdent -> Int -> GenErr BESemicolonDecl
declVarIntSemicolon lowIdent n = do
    identifier <- transLowIdent lowIdent 
    let types = BEType [] BEPrimitiveInt []
    let ident = BEDeclIdentVar identifier
    let initSc = BEInitScalar $ BEEConst $ BEConstLiteralInt $ toInteger n
    return $ BESemicolonDecl $ BEDeclInit types ident [] initSc

declVar :: LowIdent -> GenErr ()
declVar lowIdent = do
    identifier <- transLowIdent lowIdent 
    varPos <- incVarCount
    addToLocal (BEEVar identifier) (varPos-1)

transStmtBlockToSingleBlock :: StmtBlock -> GenErr SingleBlock
transStmtBlockToSingleBlock stmtblock = return $ SingleBlock [] $ StmtOrder OrderSeq stmtblock

transStmtBlock:: StmtBlock -> GenErr BECompound 
transStmtBlock x = case x of 
    StmtBlock st -> do 
        stmts <- mapM transStmt st
        return $ BECompound stmts  

transDeadLine :: DeadLine -> GenErr Integer 
transDeadLine x = case x of 
    DeadLineS dl -> return $ dl*1000*1000*1000
    DeadLineMs dl -> return $ dl*1000*1000
    DeadLineUs dl -> return $ dl*1000
    DeadLineNs dl -> return $ dl

transStmt :: Stmt -> GenErr BEStatement
transStmt x = case x of
  StmtAssign lowident assignop expr -> do
    assign <- assignStmt lowident assignop expr 
    return $ BEStatementExpression $ assign
  StmtArrayAssign lowident array assignop expr -> do
    assign <- assignArrayStmt lowident array assignop expr
    return $ BEStatementExpression $ assign
  StmtCall capident -> createBlockingProcCall capident
  StmtExpr expr -> transExpr expr >>= \e -> return $ BEStatementExpression $ e 
  StmtWork -> createWork
  StmtCompound stmtCompound -> transStmtCompound stmtCompound >>=
                                   \stc -> return $ BEStatementCompound stc 
  StmtPrint msg exprs -> transPrint msg exprs
  StmtDeSchedule -> createDeSchedule genStatusReady
  StmtChanCom chancom  -> transChanCom chancom

assignStmt :: LowIdent -> AssignOp -> Expr -> GenErr BEExp
assignStmt lowident _ expr = do
    ident <- transLowIdentToStackVar lowident
    expr0 <- transExpr expr 
    return $ BEEAssign ident BEAssign expr0

assignArrayStmt :: LowIdent -> Array -> AssignOp -> Expr -> GenErr BEExp
assignArrayStmt lowident array _ expr = do 
    tarray <- transArray array  
    ident  <- transLowIdentToStackArray lowident $ tarray
    expr0  <- transExpr expr
    return $ BEEAssign ident BEAssign expr0

createBlockingProcCall :: CapIdent -> GenErr BEStatement 
createBlockingProcCall ident = do 
    call <- createProcCallChanAndProcDep ident
    deSched <- createDeSchedule genStatusWaiting 
    return $ BEStatementCompound $ BECompound $ [call, deSched, BEEmptyStatement]

createProcCallChanAndProcDep :: CapIdent -> GenErr BEStatement
createProcCallChanAndProcDep x = case x of 
    CapIdent name -> helpCreateNewProcChanAndProcDep name chanMapContainsProc

createWork :: GenErr BEStatement
createWork = do
  declVar genWorkVar
  var <- transLowIdentToStackVar  genWorkVar 
  let block = BECompound $ [helpCallFuncStr "work" []]
  return $ BEStatementCompound $ BECompound $ [helpCreateFor var 0 genWorkLim block]

transPrint :: String -> [Expr] -> GenErr BEStatement
transPrint msg exprs = do
    (str,expr) <- mapM createPrintString exprs >>= \expr -> return $ unzip expr
    return $ helpCallFuncExp genPrint 
        $ BEEConst (BEConstLiteralString (msg ++ (concat str)++"\n")) : expr
 
createPrintString :: Expr -> GenErr (String, BEExp)
createPrintString expr = do 
    ex <- transExpr expr 
    return ("%d, ", ex)

createDeSchedule :: String -> GenErr BEStatement
createDeSchedule status = do
   pc <- incPcCount
   let assign = helpAssignVarVar genStackPc (show pc)
   let newCase = helpCreateCase $ toInteger pc
   let goto = helpCreateGoto genFinalizeLabel
   let newStatus = helpAssignVarVar genProcStatus status
   return $ BEStatementCompound $ BEStatementList $ [assign, newStatus, goto, newCase]
 
transChan :: Chan -> GenErr ()
transChan x = case x of 
    ChanVar lowident -> declChan lowident
    ChanArray lowident array -> declChanArray lowident array

declChan :: LowIdent -> GenErr ()
declChan chanName = do
    chan <- declChannelGlobal chanName
    addToHeader chan

declChanArray :: LowIdent -> Array -> GenErr ()
declChanArray chanName array = do 
    chan <- declChannelArrayGlobal chanName array
    addToHeader chan 

declChannelArrayGlobal :: LowIdent -> Array ->  GenErr BEGlobal 
declChannelArrayGlobal chanIdent array = do 
    chanName <- transLowIdent chanIdent
    chanArray <- transArray array
    let types = BEType [] (BEPrimitiveStruct (BEIdent genChannel)) [BEPointerSpec]
    let ident = BEDeclIdentVar chanName
    addChanArray (BEEVar chanName) chanArray
    return $ BEGlobalDecl $ BESemicolonDecl $ BEDeclNoInit types ident [BEArraySpec chanArray]

declChannelGlobal :: LowIdent -> GenErr BEGlobal
declChannelGlobal chanIdent = do
    chanName <- transLowIdent chanIdent
    let types = BEType [] (BEPrimitiveStruct (BEIdent genChannel)) [BEPointerSpec]
    let ident = BEDeclIdentVar chanName
    addChan $ BEEVar chanName
    return $ BEGlobalDecl $ BESemicolonDecl $ BEDeclNoInit types ident [] 

transVar :: Var -> GenErr ()
transVar x = case x of
  Var lowerident  -> declVar lowerident
  VarArray lowerident array -> do
    declVar lowerident
    size <- getArraySize array 
    incVarCountWith $ fromInteger (size-1)

getArraySize :: Array -> GenErr Integer
getArraySize x = case x of
    Array expr -> case expr of
        Numb n -> return n
        _ -> lift $ Bad "Array size must be specified with a integer"

transArray :: Array -> GenErr BEExp
transArray x = case x of 
    Array expr -> case expr of 
        Numb n -> return $ BEEConst $ BEConstLiteralInt n
        ExprVar ident -> transLowIdentToDataRef ident
        _ -> lift $ Bad "Array has to be accessed by a variable or integer"

transAssignOp :: AssignOp -> GenErr BEAssignmentOp
transAssignOp x = case x of
  Assign  -> return BEAssign 

transChanCom :: ChanCom -> GenErr BEStatement
transChanCom x = case x of
  ChanComSend chan expr -> sendOnChannel chan expr >>= \se -> checkChannelStatus chan se
  ChanComRecv chan expr -> receiveOnChannel chan expr >>= \re -> checkChannelStatus chan re

sendOnChannel :: ChanExpr -> Expr -> GenErr (BECompound, BECompound)
sendOnChannel ch expr = do 
    chan <- transChanExpr ch
    ex <- transExpr expr
    addSendRecvToChan ch addSendToChanState
    let args = [ex, chan]
    let unsynch = helpCallFuncExp genChanSend args
    let synch = helpCallFuncExp genChanSendSynch args
    let regWaiting = helpCallFuncExp genChanRegWaitingSend 
                        [chan, helpCreateVar genProc]
    let sendWait = BEStatementCompound $ BECompound $ [unsynch, regWaiting]
    createChanCompound chan synch sendWait BEEmptyStatement

receiveOnChannel :: ChanExpr -> Expr -> GenErr (BECompound, BECompound)
receiveOnChannel ch ex = do 
    var <- transExpr ex
    chan <- transChanExpr ch
    addSendRecvToChan ch addRecvToChanState
    let unsynch = helpAssignVarFunc var genChanReceive [chan]
    let synch = helpAssignVarFunc var genChanReceiveSynch [chan]
    let regWaiting = helpCallFuncExp genChanRegWaitingRecv 
                        [chan, helpCreateVar genProc]
    createChanCompound chan synch regWaiting unsynch 

addSendRecvToChan :: ChanExpr -> (BEExp -> GenErr ()) -> GenErr ()
addSendRecvToChan x f =
    case x of
        ChanExpr lowIdent -> transLowIdent lowIdent >>= \var -> f (BEEVar var)
        ChanExprArray lowIdent expr -> addSendRecvArrayExprToChan lowIdent expr f

addSendRecvArrayExprToChan :: LowIdent -> Expr -> (BEExp -> GenErr ()) -> GenErr ()
addSendRecvArrayExprToChan chanIdent expr f = case expr of 
    Numb n -> transLowIdent chanIdent >>= 
                \var -> f $ BEEArray (BEEVar var) $ BEEConst $ BEConstLiteralInt n
    ExprVar ref ->  do
        chan <- transLowIdent chanIdent
        arrayRef <- transLowIdent ref
        (var, start, to) <- getCurrentLoop 
        when ((BEEVar arrayRef) /= var) $ lift $ Bad $ 
            "Channel arrays has to be refered by the loop counter " 
            ++ (show var) ++ " not " ++ (show chan)
        loopSendRecvArrayExprToChan (BEEVar chan) start (to-1) f
    _ -> lift $ Bad $ "Channel arrays must be refered to by a constant or the loop counter"

loopSendRecvArrayExprToChan :: BEExp -> Int -> Int -> (BEExp -> GenErr ()) -> GenErr ()
loopSendRecvArrayExprToChan chan start n f = do 
    f $ BEEArray chan $ BEEConst $ BEConstLiteralInt $ toInteger n
    when (start /= n) $ loopSendRecvArrayExprToChan chan start (n-1) f

checkChannelStatus :: ChanExpr -> (BECompound, BECompound) -> GenErr BEStatement 
checkChannelStatus chanName (ready, wait) = do
  chan <- transChanExpr chanName
  let lock = helpLock chan $ helpCreateVar genLock
  let ifelse = BEStatementIf $ BEIf (BEEVar (BEIdent genChanStatusVar)) ready
                $ BEElseElse wait
  let check = helpDeclAssignVarFunc genChanStatusVar genChanCheck [chan] 
                  BEPrimitiveInt
  return $ BEStatementCompound $ BECompound $ [lock] ++ check ++ [ifelse]

createChanCompound :: BEExp -> BEStatement -> BEStatement -> BEStatement -> GenErr (BECompound, BECompound) 
createChanCompound chan ready startWaiting endWaiting = do
  descheduleStmtWait <- createDeSchedule genStatusWaiting
  let unlock = helpUnlock chan $ helpCreateVar genLock
  let ifcomp = BECompound $ [ready, unlock]
  let elsecomp = BECompound $ [startWaiting, unlock, descheduleStmtWait, endWaiting]
  return $ (ifcomp, elsecomp)

transChanExpr :: ChanExpr -> GenErr BEExp
transChanExpr x = case x of
   ChanExpr lowIdent -> do 
     ident <- transLowIdent lowIdent 
     return $ BEEVar $ ident
   ChanExprArray lowIdent expr -> do
     ident <- transLowIdent lowIdent 
     ex <- transExpr expr
     return $ BEEArray (BEEVar ident) ex

transExpr :: Expr -> GenErr BEExp
transExpr x = case x of
  Plus expr0 expr  -> do 
    ex0 <- transExpr expr0
    ex <- transExpr expr
    return $ BEEPlus ex0 $ ex
  Min expr0 expr -> do 
    ex0 <- transExpr expr0
    ex <- transExpr expr
    return $ BEEMinus ex0 ex
  ExprArray expr expr1 -> transArrayExpr expr expr1
  Numb n -> return $ BEEConst $ BEConstLiteralInt n
  ExprVar ident -> transLowIdentToDataRef ident

transCapIdent :: CapIdent -> GenErr BEIdent
transCapIdent x = case x of
  CapIdent str  -> return $ BEIdent str

transLowIdentToStackVar :: LowIdent -> GenErr BEExp
transLowIdentToStackVar str = do
  ident <- transLowIdent str
  let var = BEEVar $ ident
  (level, pos) <- lookForVar var 
  return $ referToParentStack level $ BEEConst $ BEConstLiteralInt $ toInteger pos

referToParentStack :: Int -> BEExp -> BEExp
referToParentStack n var =
 if n == 0 
   then BEEArray (BEEVar (BEIdent genStackVar)) var
   else BEEPoint (BEEVar (BEIdent genParentStack)) $ (referToParentStack (n - 1) var) 
        
transLowIdentToStackArray :: LowIdent -> BEExp -> GenErr BEExp
transLowIdentToStackArray str expr = do 
    ident <- transLowIdent str
    let var = BEEVar $ ident
    (level, pos) <- lookForVar var
    return $ referToParentStack level $ BEEPlus expr $ BEEConst $ BEConstLiteralInt $ toInteger pos

transLowIdentString :: LowIdent -> GenErr String
transLowIdentString x = case x of 
    LowIdent str -> return $ str

transLowIdent :: LowIdent -> GenErr BEIdent
transLowIdent x = case x of
  LowIdent str  -> return $ BEIdent str

transLowIdentToDataRef :: LowIdent -> GenErr BEExp
transLowIdentToDataRef ident = do 
    identifier <- transLowIdent ident
    let var = BEEVar $ identifier
    (level, pos) <- lookForVar var
    return $ referToParentStack level $ BEEConst $ BEConstLiteralInt $ toInteger pos

transArrayExpr :: Expr -> Expr -> GenErr BEExp 
transArrayExpr expr expr1 =  case expr of 
      ExprVar ident -> transExpr expr1 >>= \ex -> transLowIdentToStackArray ident ex
      _ -> lift $ Bad "Array must be idetified by a variable"
   
compile :: String -> GenErr BEProgram
compile input =  let
    lexi = myLexer input
    str = pProgramSpec lexi in
    case str of 
        Bad msg -> lift $ Bad msg
        Ok bestmt -> transProgramSpec bestmt

initState :: CompState
initState = do 
  let empty = BEIdent ""
  let gs = GenState [] [] 0 0 0 0 empty ((BEEVar empty), 0, 0)
  let vs = VarState []
  let chs = ChanState M.empty
  CompState gs [vs] chs

main:: IO()
main = do 
  opts <- parseOptions 
  let Options { optInput = input, 
                optOutput = output } = opts
  str <- input
  case evalStateT (compile str) initState of
    Bad msg -> putStrLn msg
    Ok bestmt -> output $ genInclude ++ printTree bestmt
