#define _GNU_SOURCE
#include <sched.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <time.h>
#include <stdio.h>

#include "type.h"
#include "procQueue.h"
#include "stack.h"
#include "proc.h"
#include "linkedList.h"
#include "chan.h"

#define CLOCK_ID CLOCK_MONOTONIC
#define WAIT usleep(20);
#define N_WORKER 2

bool exitRuntimeStatus = false;
struct ProcQueue * procQueue;
struct ProcQueue * waitQueue;
#ifdef __DEP__
struct LinkedListProc *runQueue;
#endif //__DEP__
int nextQueue;
struct timespec startTime;
struct Worker *workers[N_WORKER];
pthread_mutex_t exitLock;

unsigned long long int clockDiff(unsigned long long dl) {
    struct timespec now;
    clock_gettime(CLOCK_ID, &now);
    return 1000*1000*1000*(now.tv_sec - startTime.tv_sec) + (now.tv_nsec - startTime.tv_nsec) + dl;
}

void initQueue(void) {
    nextQueue = 0;
    procQueue = makeProcQueue();
    waitQueue = makeProcQueue();

#ifdef __DEP__
    runQueue = linkedListProcInit();
#endif //__DEP__
    clock_gettime(CLOCK_ID, &startTime);
}

typedef void* (*threadFuncT)(void *);

void *dowork();
void initWorkers(void) {
    pthread_mutex_init(&exitLock, NULL);
    for(int i = 0; i < N_WORKER; i++) {
        workers[i] = malloc(sizeof(struct Worker));
        workers[i]->next = NULL;
        pthread_mutex_init(&(workers[i]->workerLock), NULL);
        pthread_create(&(workers[i]->thread), NULL, (threadFuncT)dowork, (void *)i);
    }
}

void waitWorkers(void){
    for(int i = 0; i < N_WORKER; i++) {
        pthread_join(workers[i]->thread, NULL);
    }
}

struct Worker *findWorker() {
    for(int i = 0; i < N_WORKER; i ++) {
        if(workers[i]->thread == pthread_self()) {
            return workers[i];
        }
    }
    return NULL;
}

void exitRuntime() {
    static int n = 0;
    pthread_mutex_lock(&exitLock);
    if (n == N_WORKER) {
        freeProcQueue(procQueue);
        procQueue = NULL;
        freeProcQueue(waitQueue);
        waitQueue = NULL;
#ifdef __DEP__
        linkedListProcFree(runQueue);
        runQueue = NULL;
#endif //__DEP__
        for(int i = 0; i < N_WORKER; i++){
            pthread_mutex_destroy(&(workers[i]->workerLock));
        }

        pthread_mutex_unlock(&exitLock);
        pthread_mutex_destroy(&exitLock);
    }
    else {
        n++;
        exitRuntimeStatus = true;
        pthread_mutex_unlock(&exitLock);
    }
    
    _exit(0);
    pthread_exit(0);
}

struct Proc *makeProc(void *procFunc, unsigned long long int deadline) {
    struct Proc *newProc = malloc(sizeof(struct Proc));

    static int pid = 0;
    newProc->pid = pid++;
    newProc->procFunc = procFunc;
    newProc->chanArg = NULL;
    newProc->stack = NULL;
    newProc->status = ProcReady;
    newProc->deadline = clockDiff(deadline);
    newProc->depProc = linkedListProcInit();
    newProc->depChan = linkedListChanInit();
    newProc->depProcData.num = 0;
    newProc->depProcData.childrenStack = linkedListStackInit();

    return newProc;
}

void freeProc(struct Proc *proc) {
    struct Stack *stack = proc->stack;
    freeStack(stack);
    linkedListProcFree(proc->depProc);
    linkedListChanFree(proc->depChan);
    linkedListStackFree(proc->depProcData.childrenStack);
    free(proc);
}

void addDepProc(struct Proc *, struct Proc*);

struct Proc *newProc (void *procFunc,
        int nVars,
        int deadline,
        struct Stack *parent,
        void (*chanDep)(struct Proc *),
        struct Proc *depProc
        ) {
    struct Proc *newProc = makeProc(procFunc,deadline);
    newProc->stack = makeStack(nVars, (nVars * sizeof(int)), parent); 

    queueReadyProc(newProc, procQueue);

    if(chanDep) {
        chanDep(newProc);
    }

    if(depProc) {
        addDepProc(newProc, depProc);
    }

    return newProc;
}

void addDepProc(struct Proc *proc, struct Proc *depProc) {
    linkedListProcQueue(proc->depProc, depProc);
    depProc->depProcData.num ++;
}

void mergeDepProcStack(struct Proc *parent) {
    if (!parent) {
        return;
    }

    struct Stack *tmp = linkedListStackDequeue(parent->depProcData.childrenStack);
    struct Stack *parentStack = parent->stack;
    struct Stack *tmpStack = makeStack(parentStack->nVars, parentStack->size, NULL);
    copyStack(tmpStack, parentStack);
    while(tmp) {
        mergeStacks(tmpStack, parentStack, tmp);
        freeStack(tmp);
        tmp = linkedListStackDequeue(parent->depProcData.childrenStack);
    }
}

void setDepProcReady(struct Proc *proc) {
    struct Proc *tmp = linkedListProcDequeue(proc->depProc);
    while(tmp) {
#ifdef __DEBUG__
        printf("Depedent proc moved to ready queue pid:%llu\n", tmp->pid);
#endif //__DEBUG__
        tmp->depProcData.num --;

        struct Stack *parent = proc->stack->parent;
        struct Stack *copy = makeStack(parent->nVars, parent->size, NULL);
        copyStack(copy, parent);
        linkedListStackQueue(tmp->depProcData.childrenStack, copy);

        if(tmp->depProcData.num == 0) {
            mergeDepProcStack(tmp);
            moveFromWaitingQueueToReadyQueue(tmp);
        }

        tmp = linkedListProcDequeue(proc->depProc);
    }
}

#ifdef __DEP__
void setCurrentRunQueue(struct Proc *proc, struct LinkedListChan *depChan) {
    struct Channel *node = linkedListChanDequeue(depChan);
    while (node) {
        struct Proc *procNode = linkedListProcDequeue(node->depProc);
        while(procNode) {
            setCurrentRunQueue(proc, procNode->depChan);
            linkedListProcQueue(runQueue, procNode);
            procNode = linkedListProcDequeue(node->depProc);
        }
        linkedListProcFree(node->depProc);
        node->depProc = NULL;
        node = linkedListChanDequeue(depChan);
    }
}

struct Proc *nextProcFromRunQueue() {
    struct Proc *tmp = linkedListProcDequeue(runQueue);
    if(!tmp) {
        tmp = dequeueReadyProc(procQueue);
        if(tmp) {
            setCurrentRunQueue(tmp, tmp->depChan);
        }
    }
    else {
        removeWaitingProcFromQueue(waitQueue, tmp->pid);
    }
    return tmp;
}
#endif //__DEP__

void moveFromWaitingQueueToReadyQueue(struct Proc *proc) {
#ifdef __DEBUG__
    printf("Moved from waiting to ready pid: %llu\n", proc->pid);
#endif //__DEBUG__
    proc->status = ProcReady;
    removeReadyProcFromQueue(waitQueue, proc->deadline, proc->pid);
    queueReadyProc(proc, procQueue);
}

void procWaiting(struct Proc *current) {
    current->status = ProcWaiting;
#ifdef __DEBUG__
    printf("Moved to Waiting queue pid %llu\n", current->pid);
#endif //__DEBUG__
    queueWaitingProc(current, waitQueue); 
}

void procReady(struct Proc *current) {
    current->status = ProcReady;
#ifdef __DEP
    linkedListProcQueue(runQueue, current);
#endif //__DEP__
    queueReadyProc(current, procQueue); 
}

void procDone(struct Proc *current) {
#ifdef __DEBUG__
    printf("Done pid %llu dl %llu\n", current->pid, current->deadline);
#endif //__DEBUG__
#ifdef __DONE_MSG__ 
    unsigned long long int now = clockDiff(0);
    long long int leftToDeadline = current->deadline - now;
    printf("Done pid %llu deadline %llu now %llu diff%lld\n", current->pid,
            current->deadline, now, leftToDeadline);
#endif //__DONE_MSG__    
#ifdef __DEBUG_QUEUE__    
    printf("Done pid %llu dl %llu by %d\n", current->pid, current->deadline, (int)pthread_self());
    printf("----------READY---------\n");
    rbPrintTree(procQueue->root->rb_node);
    printf("-----------WAIT---------\n");
    rbPrintTree(waitQueue->root->rb_node);
    printf("\n");
#endif //__DEBUG_QUEUE__
    current->status = ProcDone;
    setDepProcReady(current);
    removeReadyProcFromQueue(procQueue, current->deadline, current->pid);
    removeWaitingProcFromQueue(waitQueue, current->pid);
}

void procMainDone(struct Proc *current) {
    exitRuntime();
}

//struct timespec contextSwitchTimeReady;
void checkStatus(struct Proc *current) {
#ifdef __DEBUG__
    printf("Process executing pid %llu PC %d dl %llu by worker %d \n", current->pid, current->stack->pc
            , current->deadline, (int)pthread_self()); 
#endif //__DEBUG__

    enum ProcStatus status = current->status;
    if(status == ProcReady) {
        status = (current->procFunc)(current->stack, current);
        switch(status) {
            case ProcWaiting:
                procWaiting(current);
                break;
            case ProcReady:
              procReady(current);
                break;
            case ProcRunning:
                break;
            case ProcDone:
                procDone(current);
                break;
            case ProcMainDone:
                procMainDone(current);
                break;
        }
    }
}

void *dowork(int nr) {
    //Sets the sched_affinity binding for the thread
    cpu_set_t mask;
    CPU_ZERO(&mask);
    CPU_SET(nr, &mask);
    
    if(sched_setaffinity(0, sizeof(mask), &mask) == -1) {
        printf("Warning: CPU affinity is not set\n");
    }
    
    while(1) {
       if(exitRuntimeStatus) {
            exitRuntime();
       }

       if(!procQueue || !waitQueue) {
           WAIT;
       }
       
#ifdef __DEP__
       struct Proc *current = nextProcFromRunQueue();
#else 
       struct Proc *current = dequeueReadyProc(procQueue);
#endif //__DEP__

       if(current) {
           checkStatus(current);
       }
       else {
            WAIT;
       }
    }
}
