#include <stdlib.h>
#include <stdio.h>

#include <rbtree/rbtree.h>

#include "rbQueue.h"

struct rb_root active = RB_ROOT;


void rbPrintTree(struct rb_node *node) {

    if(!node){
        printf("Empty tree\n");
        return;
    }
       
    if(node->rb_right) {
        rbPrintTree(node->rb_right);
    }
    
    if(node->rb_left) {
        rbPrintTree(node->rb_left);
    }
       
    struct rbTask *task = rb_entry(node, struct rbTask, node);
     printf("Tree: Node pid:%lld deadline:%lld \n", task->pid, (task->proc)->deadline);
    return;

}

void rbPrint(struct rbTask *task) {
    if(task && task->proc)
        printf("Task id %lld \n Deadline %lld \n", task->pid, (task->proc)->deadline);
    else
       printf("Empty task\n");
}

long long int rbCompareDeadline(struct rbTask *task,
              unsigned long long deadline,
              unsigned long long pid)
{
    if(task->proc->deadline != deadline) {
        return deadline - task->proc->deadline;
    }
    else if(task->pid != pid){
        return deadline + pid - task->proc->pid;
    }
    return 0;
}

long long int rbComparePid(struct rbTask *task, 
        unsigned long long deadline,
        unsigned long long pid)
{
    if(task->pid != pid) {
        return pid - task->pid;
    }
    return 0;
}

struct rbTask *rbSearch (
        struct rb_root *root, unsigned long long num, unsigned long long num2,
        long long int (*fp)(struct rbTask *, unsigned long long, unsigned long long))
{
    long long int compare = 0;
    struct rb_node *node = root->rb_node;
    struct rbTask *task;
    
   // printf("\n Searching for pid %llu dl %llu by %d\n", num2, num, (int)pthread_self());
    while(node) {
        task = container_of(node, struct rbTask, node);
        compare = (*fp)(task, num, num2);

        if(compare == 0){
     //       printf("Search Found %llu \n", task->pid);
            return task;
        }
        if(compare < 0) {
       //     printf("Search Left pid %llu compare %lld\n", task->pid, compare);
            node = node -> rb_left;
        }
        else if (compare > 0) {
        //    printf("Search Right pid %llu compare %lld\n", task->pid, compare);
            node = node -> rb_right;
        }
    }
    return NULL;
}

struct rbTask *rbSearchPid(struct rb_root *root, unsigned long long pid) {
   return rbSearch(root, 0, pid, &rbComparePid);
}

struct rbTask *rbSearchDeadline(struct rb_root *root, unsigned long long dl, unsigned long long pid) {
    return rbSearch(root, dl, pid, &rbCompareDeadline);
}

struct rbTask *rbGetLowestDeadline(struct rb_root *root, unsigned long long n) {
    struct rb_node *node = root->rb_node;
    struct rbTask *task;

#ifdef __DEBUG_QUEUE__
    rbPrintTree(node);
#endif //__DEBUG_QUEUE__


    while(node) {
        task = container_of(node, struct rbTask, node);
        if(node -> rb_left){
            node = node -> rb_left;
        }
        else {
#ifdef __DEBUG__
            printf("Lowest Deadline task pid %llu dl %llu\n",
                    task->proc->pid, task->proc->deadline);
#endif //__DEBUG__
            return task;
        }
    }
    return NULL;
}

void rbInsert(
    struct rb_root *root,
    struct rbTask *newTask,
    unsigned long long deadline,
    unsigned long long pid,
    long long int (*fp)(struct rbTask *, unsigned long long, unsigned long long
    )) {

    struct rb_node **link = &(root -> rb_node);
    struct rb_node *parent = *link;

    while (*link) {
        parent = *link;
        struct rbTask *task = rb_entry(parent, struct rbTask, node);
        
        if(newTask->pid == task->pid) {
            return;
        }
        long long int taskCompare = (*fp)(task, deadline, pid);

        if(taskCompare < 0) {
            link = &((*link) -> rb_left);
        }
        else { 
            link = &((*link) -> rb_right);
        }
    }
    
    rb_link_node(&newTask->node, parent, link);
    rb_insert_color(&newTask->node, root);
}

void rbInsertPid (struct rb_root *root, struct rbTask *newTask) {
    rbInsert(root, newTask, newTask->proc->deadline, newTask->pid, &rbComparePid);
}

void rbInsertDeadline (struct rb_root *root, struct rbTask *newTask) {
    rbInsert(root, newTask, newTask->proc->deadline, newTask->pid, &rbCompareDeadline);
}

struct Proc * rbDeletePid(struct rb_root *root, unsigned long long pid) {
   struct rbTask *task = rbSearchPid(root, pid); 
   struct Proc *tmp = NULL;
   if(task){
        rb_erase(&(task->node), root);
        tmp = task->proc;
        free(task);
   }
   return tmp;
}

struct Proc * rbDeleteDeadline(struct rb_root *root, 
                               unsigned long long deadline,
                               unsigned long long pid
) {
   struct rbTask *task = rbSearchDeadline(root, deadline, pid); 
   struct Proc *tmp = NULL;
   if(task){
     //   printf("sucessfully removed pid %llu\n", pid);
        rb_erase(&(task->node), root);
        tmp = task->proc;
        free(task);
   }
   return tmp;
}

void rbDeleteFull(struct rb_root *root, struct rbTask *delTask) {
        rb_erase(&(delTask->node), root);
        free(delTask);
}

void rbFinish(struct rb_root *root) {
    struct rb_node *node = rb_first(root);
    struct rb_node *victim;
    struct rbTask *rb;

    while(node) {
        rb = rb_entry(node, struct rbTask, node);
        victim = node;
        node = rb_next(node);
        if(victim) {
                rb_erase(victim, root);
                free(rb_entry(node, struct rbTask, node));
        }
    }
    free(root);
}
